# Copyright (C) 2021 FriendliAI

"""CLI for Serve"""

from dateutil.parser import parse
from typing import Optional

import typer

from pfcli.service import (
    ServiceType,
    GpuType,
)
from pfcli.service.client import (
    ServeClientService,
    build_client,
)
from pfcli.service.formatter import PanelFormatter, TableFormatter
from pfcli.utils import datetime_to_pretty_str


app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)

serve_panel = PanelFormatter(
    name="Overview",
    fields=['id', 'name', 'status', 'vm', 'gpu_type', 'num_gpus', 'start', 'endpoint'],
    headers=['ID', 'Name', 'Status', 'VM', 'Device', 'Device Cnt', 'Start', 'Endpoint'],
    extra_fields=['error'],
    extra_headers=['error']
)

serve_table = TableFormatter(
    name="Serves",
    fields=[
        'id',
        'name',
        'status',
        'vm',
        'gpu_type',
        'num_gpus',
        'start',
    ],
    headers=['ID', 'Name', 'Status', 'VM', 'Device', 'Device Cnt', 'Start'],
    extra_fields=['error'],
    extra_headers=['error']
)

serve_panel.add_substitution_rule("waiting", "[bold]waiting")
serve_panel.add_substitution_rule("enqueued", "[bold cyan]enqueued")
serve_panel.add_substitution_rule("running", "[bold blue]running")
serve_panel.add_substitution_rule("success", "[bold green]success")
serve_panel.add_substitution_rule("failed", "[bold red]failed")
serve_panel.add_substitution_rule("terminated", "[bold yellow]terminated")
serve_panel.add_substitution_rule("terminating", "[bold magenta]terminating")
serve_panel.add_substitution_rule("cancelling", "[bold magenta]cancelling")

serve_table.add_substitution_rule("waiting", "[bold]waiting")
serve_table.add_substitution_rule("enqueued", "[bold cyan]enqueued")
serve_table.add_substitution_rule("running", "[bold blue]running")
serve_table.add_substitution_rule("success", "[bold green]success")
serve_table.add_substitution_rule("failed", "[bold red]failed")
serve_table.add_substitution_rule("terminated", "[bold yellow]terminated")
serve_table.add_substitution_rule("terminating", "[bold magenta]terminating")
serve_table.add_substitution_rule("cancelling", "[bold magenta]cancelling")


@app.command()
def list(
    tail: Optional[int] = typer.Option(
        None,
        "--tail",
        help="The number of serve list to view at the tail"
    ),
    head: Optional[int] = typer.Option(
        None,
        "--head",
        help="The number of serve list to view at the head"
    )
):
    """List all serves,
    """
    client: ServeClientService = build_client(ServiceType.SERVE)
    serves = client.list_serves()

    for serve in serves:
        started_at = serve.get("start")
        if started_at is not None:
            start = datetime_to_pretty_str(parse(started_at))
        else:
            start = None    
        serve["start"] = start
    
    if tail is not None or head is not None:
        target_serve_list = []
        if tail:
            target_serve_list.extend(serves[:tail])
        if head:
            target_serve_list.extend(serves[-head:]) 
    else:
        target_serve_list = serves
    
    serve_table.render(target_serve_list)


@app.command()
def stop(
    serve_id: str = typer.Argument(
        ...,
        help="ID of serve to stop"
    )
):
    """Delete serve.
    """
    client: ServeClientService = build_client(ServiceType.SERVE)
    client.delete_serve(serve_id)


@app.command()
def view(
    serve_id: str = typer.Argument(
        ...,
        help="serve id to inspect detail."
    )
):
    """Show details of a serve.
    """
    client: ServeClientService = build_client(ServiceType.SERVE)
    serve = client.get_serve(serve_id)

    started_at = serve.get("start")
    if started_at is not None:
        start = datetime_to_pretty_str(parse(started_at))
    else:
        start = None    
    serve["start"] = start

    serve_panel.render([serve])


@app.command()
def create(
    checkpoint_id: str = typer.Option(
        ...,
        "--checkpoint-id",
        "-c",
        help="Checkpoint id to deploy."
    ),
    name: str = typer.Option(
        ...,
        "--name",
        "-n",
        help="The name of serve deployment."
    ),
    gpu_type: GpuType = typer.Option(
        ...,
        "--gpu-type",
        "-g",
        help="The GPU type where the serve is deployed."
    ),
    num_sessions: int = typer.Option(
        ...,
        "--num-sessions",
        "-s",
        help="The number of sessions of serve deployment."
    ),
):
    """Create a serve object by using model checkpoint.
    """
    request_data = {
        "name": name,
        "model_id": checkpoint_id,
        "gpu_type": gpu_type,
        "num_sessions": num_sessions
    }
    client: ServeClientService = build_client(ServiceType.SERVE)
    serve = client.create_serve(request_data)

    typer.secho(
        f"Serve ({serve['id']}) started successfully. Use 'pf serivce view <id>' to see the serve details.\n" \
        f"Run 'curl {serve['endpoint']}' for inference request",
        fg=typer.colors.BLUE
    )



# Copyright (C) 2022 FriendliAI

"""PeriFlow UserClient Service"""

from string import Template
from typing import List

from pfcli.service import GroupRole, ProjectRole
from pfcli.service.client.base import (
    ClientService,
    GroupRequestMixin,
    UserRequestMixin,
    safe_request,
)
from pfcli.utils import paginated_get


class UserSignUpService(ClientService):
    def sign_up(self, username: str, name: str, email: str, password: str) -> None:
        safe_request(self.bare_post, err_prefix="Failed to signup")(
            json={"username": username, "name": name, "email": email, "password": password}
        )

    def verify(self, token: str) -> None:
        safe_request(self.bare_post, err_prefix="Failed to verify")(
            path="confirm",
            json={"email_token": token}
        )


class UserClientService(ClientService, UserRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_user()
        super().__init__(template, **kwargs)

    def change_password(self, old_password: str, new_password: str) -> None:
        safe_request(self.update, err_prefix="Failed to change password.")(
            pk=self.user_id,
            path="password",
            json={
                "old_password": old_password,
                "new_password": new_password
            }
        )

    def set_group_privilege(self, pf_group_id: str, pf_user_id: str, privilege_level: GroupRole) -> None:
        safe_request(self.partial_update, err_prefix="Failed to update privilege level in group")(
            pk=pf_user_id,
            path=f"pf_group/{pf_group_id}/privilege_level",
            json={
                "privilege_level": privilege_level.value
            }
        )

    def get_project_membership(self, pf_project_id: str) -> dict:
        response = safe_request(self.retrieve, err_prefix="Failed identify member in project")(
            pk=self.user_id,
            path=f"pf_project/{pf_project_id}",
        )
        return response.json()

    def add_to_project(self, pf_user_id: str, pf_project_id: str, access_level: ProjectRole) -> None:
        safe_request(self.post, err_prefix="Failed to add user to project")(
            path=f"{pf_user_id}/pf_project/{pf_project_id}",
            json={
                "access_level": access_level.value
            }
        )

    def set_project_privilege(self, pf_user_id: str, pf_project_id: str, access_level: ProjectRole) -> None:
        safe_request(self.partial_update, err_prefix="Failed to update privilege level in project")(
            pk=pf_user_id,
            path=f"pf_project/{pf_project_id}/access_level",
            json={
                "access_level": access_level.value
            }
        )


class UserGroupClientService(ClientService, UserRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_user()
        super().__init__(template, pf_user_id=self.user_id, **kwargs)

    def get_group_info(self) -> list:
        response = safe_request(self.list, err_prefix="Failed to get my group info.")()
        return response.json()


class UserGroupProjectClientService(ClientService, UserRequestMixin, GroupRequestMixin):
    def __init__(self, template: Template, **kwargs):
        self.initialize_user()
        self.initialize_group()
        super().__init__(template, pf_user_id=self.user_id, pf_group_id=self.group_id, **kwargs)

    def list_projects(self) -> List[dict]:
        get_response_dict = safe_request(self.list, err_prefix="Failed to list projects.")
        return paginated_get(get_response_dict)



# Copyright (C) 2022 FriendliAI

"""CLI for artifact"""

import os
from pathlib import Path
from typing import Optional

import typer

from pfcli.service import ServiceType
from pfcli.service.client import build_client
from pfcli.service.client.job import JobArtifactClientService
from pfcli.utils import download_file, secho_error_and_exit


app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)


@app.command()
def download(
    job_id: int = typer.Argument(
        ...,
        help="ID of a job to download artifact"
    ),
    save_directory: Optional[Path] = typer.Option(
        None,
        '--destination',
        '-d',
        help='Destination path to save artifact files.'
    )
):
    """download artifact
    """
    if save_directory is not None and not save_directory.is_dir():
        secho_error_and_exit(f"{save_directory.name} already exist, but not a directory!")

    save_directory = save_directory or Path(os.getcwd())

    client: JobArtifactClientService = build_client(ServiceType.JOB_ARTIFACT, job_id=job_id)
    all_artifacts = client.list_artifacts()
    for i, artifact in enumerate(all_artifacts):
        artifact_id = artifact['id']
        response = client.get_artifact_download_url(artifact_id)
        url = response["url"]
        name = artifact["name"]
        typer.secho(f"Downloading files {i + 1}/{len(all_artifacts)}...")
        download_file(url, out=str(save_directory / name))



# Copyright (C) 2021 FriendliAI

"""PeriFlow Job"""

import asyncio
import re
from pathlib import Path
from typing import Dict, Generator, Optional, List, Tuple
from dateutil import parser
from dateutil.parser import parse
from datetime import datetime

import tabulate
import typer
import yaml
import ruamel.yaml
from click import Choice

from pfcli.service import JobType, ServiceType, storage_type_map_inv
from pfcli.service.client import (
    ProjectDataClientService,
    ProjectExperimentClientService,
    ProjectJobClientService,
    GroupVMConfigClientService,
    JobArtifactClientService,
    JobCheckpointClientService,
    JobClientService,
    JobTemplateClientService,
    JobWebSocketClientService,
    build_client,
)
from pfcli.service.client.project import ProjectClientService
from pfcli.service.config import build_job_configurator
from pfcli.service.formatter import PanelFormatter, TableFormatter
from pfcli.utils import (
    get_default_editor,
    open_editor,
    secho_error_and_exit,
    datetime_to_pretty_str,
    timedelta_to_pretty_str,
    utc_to_local,
    datetime_to_simple_string,
)

tabulate.PRESERVE_WHITESPACE = True

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)
template_app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)

app.add_typer(template_app, name="template", help="Manager job templates.")

job_table = TableFormatter(
    name="Jobs",
    fields=[
        'id',
        'project',
        'name',
        'status',
        'vm_config.vm_config_type.vm_instance_type.name',
        'vm_config.vm_config_type.vm_instance_type.device_type',
        'num_desired_devices',
        'data_name',
        'started_at',
        'duration',
    ],
    headers=['ID', 'Project', 'Name', 'Status', 'VM', 'Device', 'Device Cnt', 'Data', 'Start', 'Duration'],
)
job_table.apply_styling("ID", style="bold")
job_table.add_substitution_rule("waiting", "[bold]waiting")
job_table.add_substitution_rule("enqueued", "[bold cyan]enqueued")
job_table.add_substitution_rule("running", "[bold blue]running")
job_table.add_substitution_rule("success", "[bold green]success")
job_table.add_substitution_rule("failed", "[bold red]failed")
job_table.add_substitution_rule("terminated", "[bold yellow]terminated")
job_table.add_substitution_rule("terminating", "[bold magenta]terminating")
job_table.add_substitution_rule("cancelling", "[bold magenta]cancelling")
job_table.add_substitution_rule("None", "-")
job_panel = PanelFormatter(
    name="Overview",
    fields=[
        'id',
        'name',
        'status',
        'vm_config.vm_config_type.vm_instance_type.name',
        'vm_config.vm_config_type.vm_instance_type.device_type',
        'num_desired_devices',
        'data_name',
        'started_at',
        'duration',
    ],
    headers=['ID', 'Name', 'Status', 'VM', 'Device', 'Device Cnt', 'Data', 'Start', 'Duration'],
)
job_panel.add_substitution_rule("waiting", "[bold]waiting")
job_panel.add_substitution_rule("enqueued", "[bold cyan]enqueued")
job_panel.add_substitution_rule("running", "[bold blue]running")
job_panel.add_substitution_rule("success", "[bold green]success")
job_panel.add_substitution_rule("failed", "[bold red]failed")
job_panel.add_substitution_rule("terminated", "[bold yellow]terminated")
job_panel.add_substitution_rule("terminating", "[bold magenta]terminating")
job_panel.add_substitution_rule("cancelling", "[bold magenta]cancelling")
job_panel.add_substitution_rule("None", "-")
ckpt_table = TableFormatter(
    name="Checkpoints",
    fields=['id', 'vendor', 'region', 'iteration', 'model_form_category', 'created_at'],
    headers=['ID', 'Cloud', 'Region', 'Iteration', 'Format', 'Created At']
)
artifact_table = TableFormatter(
    name="Artifacts",
    fields=['id', 'name', 'path', 'mtime', 'mime_type'],
    headers=['ID', 'Name', 'Path', 'Mtime', 'Media Type']
)


def refine_config(config: dict,
                  vm_name: Optional[str],
                  num_devices: Optional[int],
                  experiment_name: Optional[str],
                  job_name: Optional[str]) -> None:
    assert "job_setting" in config

    if num_devices is not None:
        config["num_devices"] = num_devices
    else:
        assert "num_devices" in config

    if job_name is not None:
        config["name"] = job_name

    if config["job_setting"]["type"] == "custom" and "workspace" not in config["job_setting"]:
        config["job_setting"]["workspace"] = {"mount_path": "/workspace"}

    experiment_client: ProjectExperimentClientService = build_client(ServiceType.PROJECT_EXPERIMENT)
    data_client: ProjectDataClientService = build_client(ServiceType.PROJECT_DATA)
    vm_client: GroupVMConfigClientService = build_client(ServiceType.GROUP_VM_CONFIG)
    job_template_client: JobTemplateClientService = build_client(ServiceType.JOB_TEMPLATE)

    experiment_name = experiment_name or config["experiment"]
    experiment_id = experiment_client.get_id_by_name(experiment_name)
    if experiment_id is None:
        create_new = typer.confirm(
            f"Experiment with the name ({experiment_name}) is not found.\n"
            "Do you want to proceed with creating a new experiment? "
            f"If yes, a new experiment will be created with the name: {experiment_name}",
            prompt_suffix="\n>> "
        )
        if not create_new:
            typer.echo("Please run the job after creating an experiment first.")
            raise typer.Abort()
        experiment_id = experiment_client.create_experiment(experiment_name)["id"]
        typer.echo(f"A new experiment ({experiment_name}) is created.")

    del config["experiment"]
    config["experiment_id"] = experiment_id

    vm_name = vm_name or config["vm"]
    vm_config_id = vm_client.get_id_by_name(vm_name)
    if vm_config_id is None:
        secho_error_and_exit(f"VM ({vm_name}) is not found.")
    del config["vm"]
    config["vm_config_id"] = vm_config_id

    if "data" in config:
        data_name = config["data"]["name"]
        data_id = data_client.get_id_by_name(data_name)
        if data_id is None:
            secho_error_and_exit(f"Dataset ({data_name}) is not found.")
        del config["data"]["name"]
        config["data"]["id"] = data_id

    if config["job_setting"]["type"] == "custom":
        config["job_setting"]["launch_mode"] = "node"
    else:
        job_template_name = config["job_setting"]["template_name"]
        job_template_config = job_template_client.get_job_template_by_name(job_template_name)
        if job_template_config is None:
            secho_error_and_exit(f"Predefined job template ({job_template_name}) is not found.")
        del config["job_setting"]["template_name"]
        config["job_setting"]["engine_code"] = job_template_config["engine_code"]
        config["job_setting"]["model_code"] = job_template_config["model_code"]


@app.command()
def run(
    config_file: typer.FileText = typer.Option(
        ...,
        "--config-file",
        "-f",
        help="Path to configuration file"
    ),
    workspace_dir: Optional[Path] = typer.Option(
        None,
        "--workspace-dir",
        "-w",
        help="Path to workspace directory in your local file system"
    ),
    experiment_name: Optional[str] = typer.Option(
        None,
        "--experiment",
        "-e",
        help="The name of experiment. "
             "If not provided, the value in the config file will be used."
    ),
    job_name: Optional[str] = typer.Option(
       None,
       "--name",
       "-n",
       help="The name of this job. " 
             "If not provided, the value in the config file will be used."
    ),
    vm_name: Optional[str] = typer.Option(
        None,
        "--vm",
        "-v",
        help="VM type. You can check the list of VMs with `pf vm list`. "
             "If not provided, the value in the config file will be used."
    ),
    num_devices: Optional[int] = typer.Option(
        None,
        "--num-devices",
        "-d",
        help="The number of devices to use in the job"
    )
):
    """Run a job.
    """
    try:
        config: dict = yaml.safe_load(config_file)
    except yaml.YAMLError as e:
        secho_error_and_exit(f"Error occurred while parsing config file... {e}")

    refine_config(config, vm_name, num_devices, experiment_name, job_name)

    if workspace_dir is not None:
        workspace_dir = workspace_dir.resolve()  # ensure absolute path
        if not workspace_dir.exists():
            secho_error_and_exit(f"Specified workspace does not exist...")
        if not workspace_dir.is_dir():
            secho_error_and_exit(f"Specified workspace is not directory...")

    client: ProjectJobClientService = build_client(ServiceType.PROJECT_JOB)
    job_data = client.run_job(config, workspace_dir)

    typer.secho(
        f"Job ({job_data['id']}) started successfully. Use 'pf job log {job_data['id']}' to see the job logs.",
        fg=typer.colors.BLUE
    )


@app.command()
def list(
    tail: Optional[int] = typer.Option(
        None,
        "--tail",
        help="The number of job list to view at the tail"
    ),
    head: Optional[int] = typer.Option(
        None,
        "--head",
        help="The number of job list to view at the head"
    ),
    show_project_job: bool = typer.Option(
        False,
        "--project",
        "-p",
        help="Show all jobs in my project including jobs launched by other users"
    )
):
    """List all jobs.
    """
    if show_project_job:
        client: ProjectJobClientService = build_client(ServiceType.PROJECT_JOB)
    else:
        client: JobClientService = build_client(ServiceType.JOB)
    jobs = client.list_jobs()
    project_client: ProjectClientService = build_client(ServiceType.PROJECT)
    project_name_cache: Dict[str, str] = {}

    for job in jobs:
        started_at = job.get("started_at")
        finished_at = job.get("finished_at")
        if started_at is not None:
            start = datetime_to_pretty_str(parse(job["started_at"]))
        else:
            start = None
        if started_at is not None and finished_at is not None:
            duration = timedelta_to_pretty_str(parse(started_at), parse(finished_at))
        elif started_at is not None and job["status"] == "running":
            start_time = parse(started_at)
            curr_time = datetime.now(start_time.tzinfo)
            duration = timedelta_to_pretty_str(start_time, curr_time)
        else:
            duration = None

        project_id = job.get("project_id")
        if project_id is not None:
            if project_id in project_name_cache:
                job['project'] = project_name_cache[project_id]
            else:
                job['project'] = project_client.get_project(pf_project_id=project_id)['name']
                project_name_cache[project_id] = job['project']
        job['started_at'] = start
        job['duration'] = duration
        job['data_name'] = job["data_store"]['name'] if job["data_store"] is not None else None

    if tail is not None or head is not None:
        target_job_list = []
        if tail:
            target_job_list.extend(jobs[:tail])
        if head:
            target_job_list.extend(jobs[-head:]) 
    else:
        target_job_list = jobs

    job_table.render(target_job_list)


@app.command()
def stop(
    job_id: int = typer.Argument(
        ...,
        help="ID of job to stop"
    )
):
    """Termiate/cancel a running/enqueued job.
    """
    client: JobClientService = build_client(ServiceType.JOB)
    job_status = client.get_job(job_id)["status"]

    if job_status == "waiting":
        client.cancel_job(job_id)
    elif job_status in ("running", "enqueued"):
        client.terminate_job(job_id)
    else:
        secho_error_and_exit(f"No need to stop {job_status} job...")


@app.command()
def delete(
    job_id: int = typer.Argument(
        ...,
        help="ID of job to delete"
    ),
):
    """Delete a job.
    """
    do_delete = typer.confirm("Are your sure to delete job?")
    if not do_delete:
        raise typer.Abort()

    client: JobClientService = build_client(ServiceType.JOB)
    client.delete_job(job_id)

    typer.secho(f"Job ({job_id}) deleted successfully!", fg=typer.colors.BLUE)


@app.command()
def view(
    job_id: int = typer.Argument(
        ...,
        help="ID of job to view detail"
    ),
):
    """Show job detail.
    """
    job_client: JobClientService = build_client(ServiceType.JOB)
    job_checkpoint_client: JobCheckpointClientService = build_client(ServiceType.JOB_CHECKPOINT, job_id=job_id)
    job_artifact_client: JobArtifactClientService = build_client(ServiceType.JOB_ARTIFACT, job_id=job_id)

    job = job_client.get_job(job_id)
    job_checkpoints = job_checkpoint_client.list_checkpoints()
    job_artifacts = job_artifact_client.list_artifacts()

    started_at = job.get("started_at")
    finished_at = job.get("finished_at")
    if started_at is not None:
        start = datetime_to_pretty_str(parse(job["started_at"]))
    else:
        start = None
    if started_at is not None and finished_at is not None:
        duration = timedelta_to_pretty_str(parse(started_at), parse(finished_at))
    elif started_at is not None and job["status"] == "running":
        start_time = parse(started_at)
        curr_time = datetime.now(start_time.tzinfo)
        duration = timedelta_to_pretty_str(start_time, curr_time)
    else:
        duration = None

    job['started_at'] = start
    job['duration'] = duration
    job['data_name'] = job["data_store"]['name'] if job["data_store"] is not None else None

    checkpoint_list = []
    for checkpoint in reversed(job_checkpoints):
        checkpoint['created_at'] = datetime_to_pretty_str(parse(checkpoint["created_at"]), long_list=True)
        checkpoint['vendor'] = storage_type_map_inv[checkpoint['vendor']].value
        checkpoint_list.append(checkpoint)

    job_panel.render([job], show_detail=True)
    ckpt_table.render(checkpoint_list)
    artifact_table.render(job_artifacts)


@template_app.command("create")
def template_create(
    save_path: typer.FileTextWrite = typer.Option(
        ...,
        "--save-path",
        "-s",
        help="Path to save job YAML configruation file."
    )
):
    """Create a job configuration YAML file
    """
    job_type = typer.prompt(
        "What kind of job do you want?\n",
        type=Choice([ e.value for e in JobType ]),
        prompt_suffix="\n>> "
    )
    configurator = build_job_configurator(job_type)
    yaml_str = configurator.render()

    yaml = ruamel.yaml.YAML()
    code = yaml.load(yaml_str)
    yaml.dump(code, save_path)

    continue_edit = typer.confirm(
        f"Do you want to open an editor to configure the job YAML file? (default editor: {get_default_editor()})",
        prompt_suffix="\n>> "
    )
    if continue_edit:
        open_editor(save_path.name)


def _split_machine_ids(value: Optional[str]) -> Optional[List[int]]:
    if value is None:
        return value
    try:
        return [ int(machine_id) for machine_id in value.split(",") ]
    except ValueError:
        secho_error_and_exit("Machine index should be integer. (e.g., --machine 0,1,2)")


def _format_log_string(log_record: dict,
                       show_time: bool,
                       show_machine_id: bool,
                       use_style: bool = True) -> Generator[Tuple[str, bool], None, None]:
    timestamp_str = f"⏰ {datetime_to_simple_string(utc_to_local(parser.parse(log_record['timestamp'])))} "
    node_rank = log_record['node_rank']
    node_rank_str = "📈 PF " if node_rank == -1 else f"💻 #{node_rank} "

    if use_style:
        timestamp_str = typer.style(timestamp_str, fg=typer.colors.BLUE)
        node_rank_str = typer.style(node_rank_str, fg=typer.colors.GREEN)

    lines = [x for x in re.split(r'(\n|\r)', log_record['content']) if x]

    job_finished = False
    for line in lines:
        if line in ('\n', '\r'):
            yield line, job_finished
        else:
            job_finished = (line in ("Job completed successfully.", "Job failed.") and node_rank == -1)
            if show_machine_id:
                line = node_rank_str + line
            if show_time:
                line = timestamp_str + line
            yield line, job_finished


async def monitor_logs(job_id: int,
                       log_types: Optional[List[str]],
                       machines: Optional[List[int]],
                       show_time: bool,
                       show_machine_id: bool):
    ws_client: JobWebSocketClientService = build_client(ServiceType.JOB_WS)

    job_finished = False
    async with ws_client.open_connection(job_id, log_types, machines):
        async for response in ws_client:
            for line, job_finished in _format_log_string(response, show_time, show_machine_id):
                typer.echo(line, nl=False)
            if job_finished:
                return


# TODO: Implement since/until if necessary
@app.command()
def log(
    job_id: int = typer.Argument(
        ...,
        help="ID of job to view log"
    ),
    num_records: int = typer.Option(
        100,
        "--num-records",
        "-n",
        help="The number of recent records to view"
    ),
    content: Optional[str] = typer.Option(
        None,
        "--content",
        "-c",
        help="Filter logs by content"
    ),
    machines: str = typer.Option(
        None,
        "--machine",
        "-m",
        callback=_split_machine_ids,
        help="Filter logs by machine ID. Comma-separated indices of machine to print logs (e.g., 0,1,2,3). "
             "By default, it will print logs from all machines."
    ),
    head: bool = typer.Option(
        False,
        "--head",
        help="View logs from the oldest one"
    ),
    export_path: Optional[Path] = typer.Option(
        None,
        "--export-path",
        "-e",
        help="Path to export logs"
    ),
    follow: bool = typer.Option(
        False,
        "--follow",
        "-f",
        help="Follow logs"
    ),
    show_time: bool = typer.Option(
        False,
        "--timestamp",
        "-t",
        help="Print logs with timestamp"
    ),
    show_machine_id: bool = typer.Option(
        False,
        "--show-machine-id",
        help="Print logs with machine index"
    )
):
    """Show job logs.
    """
    if num_records <= 0 or num_records > 10000:
        secho_error_and_exit("'num_records' should be a positive integer, equal or smaller than 10000")

    if head and follow:
        secho_error_and_exit("'follow' cannot be set in 'head' mode")

    if export_path is not None and follow:
        secho_error_and_exit("'follow' cannot be set when 'export_path' is given")

    client: JobClientService = build_client(ServiceType.JOB)
    logs = client.get_text_logs(job_id, num_records, head, None, machines, content)

    job_finished = False
    if export_path is not None:
        with export_path.open("w", encoding="utf-8") as export_file:
            for record in logs:
                for line, _ in _format_log_string(record, show_time, show_machine_id, use_style=False):
                    export_file.write(line)
    else:
        for record in logs:
            for line, job_finished in _format_log_string(record, show_time, show_machine_id):
                typer.echo(line, nl=False)

    if not job_finished and follow:
        try:
            # Subscribe job log
            asyncio.run(
                monitor_logs(job_id, None, machines, show_time, show_machine_id)
            )
        except KeyboardInterrupt:
            secho_error_and_exit(f"Keyboard Interrupt...", color=typer.colors.MAGENTA)



# Copyright (C) 2021 FriendliAI

"""PeriFlow CLI"""

import requests
from requests import HTTPError

import typer

from pfcli import (
    artifact,
    billing,
    checkpoint,
    credential,
    datastore,
    experiment,
    group,
    job,
    project,
    serve,
    vm,
)
from pfcli.service import ServiceType
from pfcli.service.auth import TokenType, update_token
from pfcli.service.client import UserClientService, UserSignUpService, build_client
from pfcli.service.formatter import PanelFormatter
from pfcli.utils import get_uri, secho_error_and_exit

app = typer.Typer(
    help="Welcome to PeriFlow 🤗",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)
app.add_typer(credential.app, name="credential", help="Manage credentials")
app.add_typer(job.app, name="job", help="Manage jobs")
app.add_typer(checkpoint.app, name="checkpoint", help="Manage checkpoints")
app.add_typer(datastore.app, name="datastore", help="Manage datasets")
app.add_typer(vm.app, name="vm", help="Manage VMs")
app.add_typer(experiment.app, name="experiment", help="Manage experiments")
app.add_typer(serve.app, name="serve", help="Manage serves")
app.add_typer(project.app, name="project", help="Manage projects")
app.add_typer(group.app, name="org", help="Manage organizations")
app.add_typer(artifact.app, name="artifact", help="Manager artifacts")
app.add_typer(billing.app, name="billing", help="Manage billing")


user_panel_formatter = PanelFormatter(
    name="My Info",
    fields=['name', 'username', 'email'],
    headers=["Name", "Username", "Email"]
)

@app.command(help="Sign up to PerfiFlow")
def signup(
    username: str = typer.Option(..., prompt="Enter Username"),
    name: str = typer.Option(..., prompt="Enter Name"),
    email: str = typer.Option(..., prompt="Enter Email"),
    password: str = typer.Option(..., prompt="Enter Password", hide_input=True),
    confirm_password: str = typer.Option(..., prompt="Enter the password again (confirmation)", hide_input=True)
):
    if password != confirm_password:
        secho_error_and_exit("Passwords did not match.")

    client: UserSignUpService = build_client(ServiceType.SIGNUP)
    client.sign_up(username, name, email, password)

    typer.echo(f"\n\nWe just sent a verification code over to {email}")
    typer.run(_verify)


@app.command(help="Show my user info")
def whoami():
    client: UserClientService = build_client(ServiceType.USER)
    info = client.get_current_userinfo()
    user_panel_formatter.render([info])


@app.command(help="Sign in PeriFlow")
def login(
    username: str = typer.Option(..., prompt="Enter Username"),
    password: str = typer.Option(..., prompt="Enter Password", hide_input=True)
):
    r = requests.post(get_uri("token/"), data={"username": username, "password": password})
    try:
        r.raise_for_status()
        update_token(token_type=TokenType.ACCESS, token=r.json()["access_token"])
        update_token(token_type=TokenType.REFRESH, token=r.json()["refresh_token"])

        typer.echo("\n\nLogin success!")
        typer.echo("Welcome back to...")
        typer.echo(" _____          _  _____ _")
        typer.echo("|  __ \___ _ __(_)|  ___| | _____      __")
        typer.echo("|  ___/ _ \ '__| || |__ | |/ _ \ \ /\ / /")
        typer.echo("| |  |  __/ |  | ||  __|| | (_) | V  V / ")
        typer.echo("|_|   \___|_|  |_||_|   |_|\___/ \_/\_/  ")
        typer.echo("\n\n")
    except HTTPError:
        secho_error_and_exit("Login failed... Please check your username and password.")


@app.command(help="Change your password")
def passwd(
    old_password: str = typer.Option(..., prompt="Enter your current password", hide_input=True),
    new_password: str = typer.Option(..., prompt="Enter your new password", hide_input=True),
    confirm_password: str = typer.Option(..., prompt="Enter the new password again (confirmation)", hide_input=True)
):
    if old_password == new_password:
        secho_error_and_exit("The current password is the same with the new password.")
    if new_password != confirm_password:
        secho_error_and_exit("Passwords did not match.")
    client: UserClientService = build_client(ServiceType.USER)
    client.change_password(old_password, new_password)

    typer.secho("Password is changed successfully!", fg=typer.colors.BLUE)


def _verify(_, token: str = typer.Option(..., prompt="Enter Code")):
    client: UserSignUpService = build_client(ServiceType.SIGNUP)
    client.verify(token)

    typer.echo("\n\nVerified!")
    typer.echo("Sign up success! Please sign in.")



# Copyright (C) 2021 FriendliAI

"""PeriFlow CLI Shared Utilities"""

import os
import zipfile
from concurrent.futures import ThreadPoolExecutor, wait, FIRST_EXCEPTION
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from dateutil.tz import tzlocal
from pathlib import Path
from subprocess import CalledProcessError, check_call
from typing import Callable, Optional, List, Dict
from urllib.parse import urljoin

import pathspec
import typer
import requests
from requests.exceptions import HTTPError
from requests.models import Response
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from pfcli.service import (
    CloudType,
    StorageType,
    storage_region_map,
    cloud_region_map
)

# Variables
periflow_directory = Path.home() / ".periflow"
periflow_api_server = "https://api-staging.friendli.ai/api/"
periflow_ws_server = "wss://api-ws-staging.friendli.ai/ws/"
periflow_discuss_url = "https://discuss-staging.friendli.ai/"
periflow_mr_server = "https://pfmodelregistry-staging.friendli.ai/"
periflow_serve_server = "http://0.0.0.0:8000/"
periflow_auth_server = "https://pfauth-staging.friendli.ai/"
periflow_meter_server = "https://pfmeter-staging.friendli.ai/"


def get_periflow_directory() -> Path:
    periflow_directory.mkdir(exist_ok=True)
    return periflow_directory


def datetime_to_pretty_str(past: Optional[datetime], long_list: bool = False):
    cur = datetime.now().astimezone()
    delta = cur - past
    if long_list:
        if delta < timedelta(minutes=1):
            return f'{delta.seconds % 60}s ago'
        if delta < timedelta(hours=1):
            return f'{round((delta.seconds % 3600) / 60)}m {delta.seconds % 60}s ago'
        elif delta < timedelta(days=1):
            return f'{delta.seconds // 3600}h {round((delta.seconds % 3600) / 60)}m {delta.seconds % 60}s ago'
        elif delta < timedelta(days=3):
            return f'{delta.days}d {delta.seconds // 3600}h ' \
                   f'{round((delta.seconds % 3600) / 60)}m ago'
        else:
            return past.astimezone(tz=cur.tzinfo).strftime("%Y-%m-%d %H:%M:%S")
    else:
        if delta < timedelta(hours=1):
            return f'{round((delta.seconds % 3600) / 60)} mins ago'
        elif delta < timedelta(days=1):
            return f'{round(delta.seconds / 3600)} hours ago'
        else:
            return f'{delta.days + round(delta.seconds / (3600 * 24))} days ago'


def timedelta_to_pretty_str(start: datetime, finish: datetime, long_list: bool = False):
    delta = finish - start
    if long_list:
        if delta < timedelta(minutes=1):
            return f'{(delta.seconds % 60)}s'
        if delta < timedelta(hours=1):
            return f'{(delta.seconds % 3600) // 60}m {(delta.seconds % 60)}s'
        elif delta < timedelta(days=1):
            return f'{delta.seconds // 3600}h {(delta.seconds % 3600) // 60}m {(delta.seconds % 60)}s'
        else:
            return f'{delta.days}d {delta.seconds // 3600}h ' \
                   f'{(delta.seconds % 3600) // 60}m {delta.seconds % 60}s'
    else:
        if delta < timedelta(hours=1):
            return f'{round((delta.seconds % 3600) / 60)} mins'
        elif delta < timedelta(days=1):
            return f'{round(delta.seconds / 3600)} hours'
        else:
            return f'{delta.days + round(delta.seconds / (3600 * 24))} days'


def get_auth_uri(path: str) -> str:
    return urljoin(periflow_auth_server, path)


def get_uri(path: str) -> str:
    return urljoin(periflow_api_server, path)


def get_wss_uri(path: str) -> str:
    return urljoin(periflow_ws_server, path)


def get_pfs_uri(path: str) -> str:
    return urljoin(periflow_serve_server, path)


def get_mr_uri(path: str) -> str:
    return urljoin(periflow_mr_server, path)


def get_meter_uri(path: str) -> str:
    return urljoin(periflow_meter_server, path)


def secho_error_and_exit(text: str, color: str = typer.colors.RED):
    typer.secho(text, err=True, fg=color)
    raise typer.Exit(1)


def get_remaining_terminal_columns(occupied: int) -> int:
    return os.get_terminal_size().columns - occupied


def utc_to_local(dt: datetime) -> datetime:
    return dt.replace(tzinfo=timezone.utc).astimezone(tz=None)


def datetime_to_simple_string(dt: datetime) -> str:
    return dt.strftime("%Y-%m-%d %H:%M:%S")


@contextmanager
def zip_dir(base_path: Path, target_files: List[Path], zip_path: Path):
    typer.secho("Preparing workspace directory...", fg=typer.colors.MAGENTA)
    with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_STORED) as zip_file:
        for file in target_files:
            zip_file.write(file, file.relative_to(base_path.parent))
    typer.secho("Uploading workspace directory...", fg=typer.colors.MAGENTA)
    try:
        yield zip_path.open("rb")
    finally:
        zip_path.unlink()


def open_editor(path: str, editor: Optional[str] = None):
    default_editor = editor or get_default_editor()
    try:
        check_call([default_editor, path])
    except CalledProcessError:
        typer.secho("", fg=typer.colors.RED)


def get_default_editor() -> str:
    return os.environ.get("PERIFLOW_CLI_EDITOR", "vim")


def validate_storage_region(vendor: StorageType, region: str):
    available_regions = storage_region_map[vendor]
    if region not in available_regions:
        secho_error_and_exit(
            f"'{region}' is not supported region for {vendor}. Please choose another one in {available_regions}."
        )


def validate_cloud_region(vendor: CloudType, region: str):
    available_regions = cloud_region_map[vendor]
    if region not in available_regions:
        secho_error_and_exit(
            f"'{region}' is not supported region for {vendor}. Please choose another one in {available_regions}."
        )


def get_workspace_files(dir_path: Path) -> List[Path]:
    ignore_file = dir_path / ".pfignore"
    all_files = set(x for x in dir_path.rglob("*") if x.is_file() and x != ignore_file)

    if not ignore_file.exists():
        return list(all_files)

    with open(ignore_file, "r", encoding="utf-8") as f:
        ignore_patterns = f.read()
    spec = pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, ignore_patterns.splitlines())
    matched_files = set(dir_path / x for x in spec.match_tree_files(dir_path))
    return list(all_files.difference(matched_files))


def _upload_file(file_path: str, url: str, ctx: tqdm):
    try:
        with open(file_path, 'rb') as f:
            wrapped_object = CallbackIOWrapper(ctx.update, f, 'read')
            requests.put(url, data=wrapped_object)
    except FileNotFoundError:
        secho_error_and_exit(f"{file_path} is not found.")


def _get_total_file_size(file_paths: List[str]) -> int:
    return sum([os.stat(file_path).st_size for file_path in file_paths])


def storage_path_to_local_path(storage_path: str, source_path: Path, expand: bool) -> str:
    if source_path.is_file():
        return str(source_path)

    if expand:
        return str(source_path / Path(storage_path))

    return str(source_path / Path(storage_path.split('/', 1)[1]))


def upload_files(url_dicts: List[Dict[str, str]], source_path: Path, expand: bool) -> None:
    local_paths = [ storage_path_to_local_path(url_info['path'], source_path, expand) for url_info in url_dicts ]
    total_size = _get_total_file_size(local_paths)
    upload_urls = [ url_info['upload_url'] for url_info in url_dicts ]

    with tqdm(total=total_size, unit='B', unit_scale=True, unit_divisor=1024) as t:
        with ThreadPoolExecutor() as executor:
            futs = [
                executor.submit(
                    _upload_file, local_path, upload_url, t
                ) for (local_path, upload_url) in zip(local_paths, upload_urls)
            ]
            wait(futs, return_when=FIRST_EXCEPTION)


def get_file_info(storage_path: str, source_path: Path, expand: bool) -> dict:
    loacl_path = storage_path_to_local_path(storage_path, source_path, expand)
    return {
        'name': os.path.basename(storage_path),
        'path': storage_path,
        'mtime': datetime.fromtimestamp(os.stat(loacl_path).st_mtime, tz=tzlocal()).isoformat(),
        'size': os.stat(loacl_path).st_size,
    }


def get_content_size(url: str) -> int:
    response = requests.get(url, stream=True)
    if response.status_code != 200:
        secho_error_and_exit("Failed to download (invalid url)")
    return int(response.headers['Content-Length'])


def download_range(url: str, start: int, end: int, output: str, ctx: tqdm) -> None:
    headers = {'Range': f'bytes={start}-{end}'}
    response = requests.get(url, headers=headers, stream=True)

    with open(output, 'wb') as f:
        wrapped_object = CallbackIOWrapper(ctx.update, f, 'write')
        for part in response.iter_content(1024):
            wrapped_object.write(part)


def download_file_simple(url: str, out: str, content_length: int) -> None:
    response = requests.get(url, stream=True)
    with tqdm.wrapattr(open(out, "wb"), "write", miniters=1, total=content_length) as fout:
        for chunk in response.iter_content(chunk_size=4096):
            fout.write(chunk)


def download_file_parallel(url: str, out: str, content_length: int, chunk_size: int = 1024 * 1024 * 4) -> None:
    chunks = range(0, content_length, chunk_size)

    temp_out_prefix = os.path.join(os.path.dirname(out), f'.{os.path.basename(out)}')

    try:
        with tqdm(total=content_length, unit='B', unit_scale=True, unit_divisor=1024) as t:
            with ThreadPoolExecutor() as executor:
                futs = [
                    executor.submit(
                        download_range, url, start, start + chunk_size - 1, f'{temp_out_prefix}.part{i}', t
                    ) for i, start in enumerate(chunks)
                ]
                wait(futs, return_when=FIRST_EXCEPTION)

        # Merge partitioned files
        with open(out, 'wb') as f:
            for i in range(len(chunks)):
                chunk_path = f'{temp_out_prefix}.part{i}'
                with open(chunk_path, 'rb') as chunk_f:
                    f.write(chunk_f.read())

                os.remove(chunk_path)
    finally:
        # Clean up zombie temporary partitioned files
        for i in range(len(chunks)):
            chunk_path = f'{temp_out_prefix}.part{i}'
            if os.path.isfile(chunk_path):
                os.remove(chunk_path)


def download_file(url: str, out: str) -> None:
    file_size = get_content_size(url)

    # Create directory if not exists
    dirpath = os.path.dirname(out)
    os.makedirs(dirpath, exist_ok=True)

    if file_size >= 16 * 1024 * 1024:
        download_file_parallel(url, out, file_size)
    else:
        download_file_simple(url, out, file_size)


def decode_http_err(exc: HTTPError) -> str:
    try:
        if exc.response.status_code == 500:
            error_str = f"Internal Server Error: Please contact to system admin via {periflow_discuss_url}"
        elif exc.response.status_code == 404:
            error_str = "Not Found: The requested resource is not found. Please check it again. " \
                        f"If you cannot find out why this error occurs, please visit {periflow_discuss_url}."
        else:
            response = exc.response
            detail_json = response.json()
            if 'detail' in detail_json:
                error_str = f"Error Code: {response.status_code}\nDetail: {detail_json['detail']}"
            elif 'error_description' in detail_json:
                error_str = f"Error Code: {response.status_code}\nDetail: {detail_json['error_description']}"
            else:
                error_str = f"Error Code: {response.status_code}"
    except ValueError:
        error_str = exc.response.content.decode()

    return error_str


def paginated_get(response_getter: Callable[..., Response], path: Optional[str] = None, **params) -> List[dict]:
    """Pagination listing
    """
    response_dict = response_getter(path=path, params={**params}).json()
    items = response_dict["results"]
    next_cursor = response_dict["next_cursor"]

    while next_cursor is not None:
        response_dict = response_getter(
            path=path,
            params={
                **params,
                "cursor": next_cursor
            }
        ).json()
        items.extend(response_dict["results"])
        next_cursor = response_dict["next_cursor"]

    return items



# Copyright (C) 2022 FriendliAI

"""Context (Organization / Project) managing"""

import uuid
from typing import Optional

from pfcli.utils import get_periflow_directory


org_context_path = get_periflow_directory() / "organization"
project_context_path = get_periflow_directory() / "project"


def get_current_group_id() -> Optional[uuid.UUID]:
    if not org_context_path.exists():
        return None

    with open(org_context_path, "r", encoding="utf-8") as f:
        group_id = uuid.UUID(f.read())
        return group_id


def set_current_group_id(pf_group_id: uuid.UUID):
    with open(org_context_path, "w", encoding="utf-8") as f:
        f.write(str(pf_group_id))


def get_current_project_id() -> Optional[uuid.UUID]:
    if not project_context_path.exists():
        return None

    with open(project_context_path, "r", encoding="utf-8") as f:
        project_id = uuid.UUID(f.read())
        return project_id


def set_current_project_id(pf_project_id: uuid.UUID):
    with open(project_context_path, "w", encoding="utf-8") as f:
        f.write(str(pf_project_id))



# Copyright (C) 2022 FriendliAI

"""PeriFlow CredentialClient Service"""


from typing import Optional

from pfcli.service import CredType, cred_type_map
from pfcli.service.client.base import ClientService, T, safe_request


class CredentialClientService(ClientService):
    def get_credential(self, credential_id: T) -> dict:
        response = safe_request(self.retrieve, err_prefix="Credential is not found.")(
            pk=credential_id
        )
        return response.json()

    def update_credential(self,
                          credential_id: T,
                          *,
                          name: Optional[str] = None,
                          type_version: Optional[str] = None,
                          value: Optional[dict] = None) -> dict:
        request_data = {}
        if name is not None:
            request_data['name'] = name
        if type_version is not None:
            request_data['type_version'] = type_version
        if value is not None:
            request_data['value'] = value
        response = safe_request(self.partial_update, err_prefix="Failed to updated credential")(
            pk=credential_id,
            json=request_data
        )
        return response.json()

    def delete_credential(self, credential_id: T) -> None:
        safe_request(self.delete, err_prefix="Failed to delete credential")(
            pk=credential_id
        )

class CredentialTypeClientService(ClientService):
    def get_schema_by_type(self, cred_type: CredType) -> Optional[dict]:
        type_name = cred_type_map[cred_type]
        response = safe_request(self.list, err_prefix="Failed to get credential schema.")()
        for cred_type_json in response.json():
            if cred_type_json['type_name'] == type_name:
                return cred_type_json['versions'][-1]['schema']     # use the latest version
        return None



# Copyright (C) 2022 FriendliAI

"""PeriFlow Group (Organization) CLI"""

import typer

from pfcli.context import (
    get_current_group_id,
    get_current_project_id,
    project_context_path,
    set_current_group_id
)
from pfcli.service import GroupRole, ServiceType
from pfcli.service.client import (
    GroupClientService,
    ProjectClientService,
    UserClientService,
    UserGroupClientService,
    build_client
)
from pfcli.service.formatter import PanelFormatter, TableFormatter
from pfcli.utils import secho_error_and_exit


app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)


org_table_formatter = TableFormatter(
    name="Organization",
    fields=['id', 'name'],
    headers=["ID", "Name"]
)
org_panel_formatter = PanelFormatter(
    name="Organization Detail",
    fields=['id', 'name', 'status'],
    headers=['ID', 'Name', 'Status']
)
member_table_formatter = TableFormatter(
    name="Members",
    fields=["id", "username", "name", "email", "privilege_level"],
    headers=["ID", "Username", "Name", "Email", "Role"]
)


@app.command(help="list all organizations")
def list():
    client: UserGroupClientService = build_client(ServiceType.USER_GROUP)
    orgs = client.get_group_info()
    org_table_formatter.render(orgs)


@app.command(help="create a new organization")
def create(
    name: str = typer.Argument(
        ...,
        help='Name of organization to create'
    )
):
    client: GroupClientService = build_client(ServiceType.GROUP)
    org = client.create_group(name=name)
    typer.secho(f"Organization ({name}) is created successfully!", fg=typer.colors.BLUE)
    org_panel_formatter.render(org)


@app.command(help="get current working organization")
def current():
    group_client: GroupClientService = build_client(ServiceType.GROUP)
    org_id = get_current_group_id()

    if org_id is None:
        secho_error_and_exit("working organization is not set")

    org = group_client.get_group(org_id)
    org_panel_formatter.render(org)


@app.command(help="switch working organization")
def switch(
    name: str = typer.Argument(
        ...,
        help='Name of organization to switch',
    )
):
    project_client: ProjectClientService = build_client(ServiceType.PROJECT)
    user_group_client: UserGroupClientService = build_client(ServiceType.USER_GROUP)
    orgs = user_group_client.get_group_info()

    org_id = None
    for org in orgs:
        if org["name"] == name:
            org_id = org["id"]
            break

    if org_id is None:
        secho_error_and_exit(f"No organization exists with name {name}.")

    project_id = get_current_project_id()
    if project_id is not None:
        if project_client.check_project_membership(pf_project_id=project_id):
            project_org_id = project_client.get_project(pf_project_id=project_id)["pf_group_id"]
            if project_org_id != org_id:
                project_context_path.unlink(missing_ok=True)
        else:
            project_context_path.unlink(missing_ok=True)

    set_current_group_id(org_id)
    typer.secho(f"Organization switched to {name}.", fg=typer.colors.BLUE)


@app.command(help="invite to current working organization")
def invite(
    email: str = typer.Argument(
        ...,
        help='Invitation recipient email address'
    )
):
    group_client: GroupClientService = build_client(ServiceType.GROUP)

    org = _get_current_org()

    if org['privilege_level'] != 'owner':
        secho_error_and_exit("Only the owner of the organization can invite/set-role.")

    group_client.invite_to_group(org['id'], email)
    typer.secho("Invitation Successfully Sent!")


@app.command("accept-invite", help="accept invitation")
def accept_invite(token: str = typer.Option(..., prompt="Enter Verification Code")):
    group_client: GroupClientService = build_client(ServiceType.GROUP)
    group_client.accept_invite(token)
    typer.secho("Verification Success!")


@app.command("set-role", help="set organization role of the user")
def set_role(
    username: str = typer.Argument(
        ...,
        help='Username to set role'
    ),
    role: GroupRole = typer.Argument(
        ...,
        help='Organization role'
    )
):
    user_client: UserClientService = build_client(ServiceType.USER)

    org = _get_current_org()

    if org['privilege_level'] != 'owner':
        secho_error_and_exit("Only the owner of the organization can invite/set-privilege.")

    user_id = _get_org_user_id_by_name(org['id'], username)
    user_client.set_group_privilege(org['id'], user_id, role)
    typer.secho(f"Organization role for user ({username}) successfully updated to {role.value}!")


def _get_org_user_id_by_name(org_id: str, username: str) -> str:
    group_client: GroupClientService = build_client(ServiceType.GROUP)
    users = group_client.get_user(org_id, username)
    for user in users:
        if user['username'] == username:
            return user['id']
    secho_error_and_exit(f"{username} is not a member of this organization.")


def _get_current_org() -> dict:
    user_group_client: UserGroupClientService = build_client(ServiceType.USER_GROUP)

    curr_org_id = get_current_group_id()
    if curr_org_id is None:
        secho_error_and_exit("working organization is not set")

    orgs = user_group_client.get_group_info()
    for org in orgs:
        if org['id'] == str(curr_org_id):
            return org

    # org context may be wrong
    secho_error_and_exit("Failed to identify organization... Please set organization again.")


@app.command(help="list up members in the current working organization")
def members():
    group_client: GroupClientService = build_client(ServiceType.GROUP)
    org_id = get_current_group_id()

    if org_id is None:
        secho_error_and_exit("working organization is not set")

    members = group_client.list_users(org_id)
    member_table_formatter.render(members)



# Copyright (C) 2021 FriendliAI

"""PeriFlow Datastore CLI"""

from pathlib import Path
from typing import Optional
from click import Choice

import typer
import yaml
from rich.text import Text

from pfcli.service import (
    StorageType,
    JobType,
    ServiceType,
    cred_type_map,
    cred_type_map_inv,
    storage_type_map_inv,
)
from pfcli.service.client import (
    CredentialClientService,
    DataClientService,
    ProjectDataClientService,
    build_client,
)
from pfcli.service.cloud import build_storage_helper
from pfcli.service.config import build_data_configurator
from pfcli.service.formatter import (
    JSONFormatter,
    PanelFormatter,
    TableFormatter,
    TreeFormatter,
)
from pfcli.utils import (
    secho_error_and_exit,
    upload_files,
    get_file_info,
)

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)

table_formatter = TableFormatter(
    name="Datastore",
    fields=['name', 'vendor', 'region', 'storage_name', 'active'],
    headers=['Name', 'Cloud', 'Region', 'Storage Name', 'Active'],
)
table_formatter.add_substitution_rule("True", Text("Y", style="green"))
table_formatter.add_substitution_rule("False", Text("N", style="red"))
table_formatter.add_substitution_rule("", "-")

panel_formatter = PanelFormatter(
    name="Overview",
    fields=['name', 'vendor', 'region', 'storage_name', 'active'],
    headers=['Name', 'Cloud', 'Region', 'Storage Name', 'Active'],
)
panel_formatter.add_substitution_rule("True", Text("Y", style="green"))
panel_formatter.add_substitution_rule("False", Text("N", style="red"))
panel_formatter.add_substitution_rule("", "-")

json_formatter = JSONFormatter(name="Metadata")
tree_formatter = TreeFormatter(name="Files")


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    interactive: bool = typer.Option(
        False,
        '--interactive',
        help="Use interactive mode."
    )
):
    supported_command = ('create',)
    if not interactive:
        return

    if ctx.invoked_subcommand not in supported_command:
        secho_error_and_exit(f"'--interactive' option is not supported for '{ctx.invoked_subcommand}' command.")

    if ctx.invoked_subcommand == 'create':
        job_type = typer.prompt(
            "What kind job would you like to create a datastore for?\n",
            type=Choice([e.value for e in JobType]),
            prompt_suffix="\n>> "
        )
        configurator = build_data_configurator(job_type)
        name, cloud, region, storage_name, credential_id, metadata, files = configurator.render()

        client: ProjectDataClientService = build_client(ServiceType.PROJECT_DATA)
        datastore = client.create_datastore(name, cloud, region, storage_name, credential_id, metadata, files, True)

        typer.secho("Datastore created successfully!", fg=typer.colors.BLUE)

        datastore['vendor'] = storage_type_map_inv[datastore['vendor']].value
        panel_formatter.render([datastore], show_detail=True)
        exit(0)


@app.command()
def list():
    """List datasets in datastore.
    """
    client: ProjectDataClientService = build_client(ServiceType.PROJECT_DATA)
    datastores = client.list_datastores()
    for datastore in datastores:
        datastore['vendor'] = storage_type_map_inv[datastore['vendor']].value
    table_formatter.render(datastores)


@app.command()
def view(
    name: str = typer.Argument(
        ...,
        help="ID or name of datastore to see detail info."
    )
):
    """View the detail of a dataset.
    """
    project_client: ProjectDataClientService = build_client(ServiceType.PROJECT_DATA)
    client: DataClientService = build_client(ServiceType.DATA)

    datastore_id = project_client.get_id_by_name(name)
    datastore = client.get_datastore(datastore_id)
    datastore['vendor'] = storage_type_map_inv[datastore['vendor']].value
    panel_formatter.render([datastore], show_detail=True)
    tree_formatter.render(datastore['files'])
    json_formatter.render(datastore['metadata'])


@app.command()
def create(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of your datastore to create."
    ),
    cloud: Optional[StorageType] = typer.Option(
        ...,
        '--cloud',
        '-c',
        help="Name of cloud storage vendor where your dataset is uploaded."
    ),
    region: Optional[str] = typer.Option(
        ...,
        '--region',
        '-r',
        help="Cloud storage region where your dataset is uploaded."
    ),
    storage_name: Optional[str] = typer.Option(
        ...,
        '--storage-name',
        '-s',
        help="The name of cloud storage where your dataset is uploaded."
    ),
    storage_path: Optional[str] = typer.Option(
        None,
        "--storage-path",
        "-p",
        help="File or direcotry path of cloud storage. The root of the storage will be used by default."
    ),
    credential_id: Optional[str] = typer.Option(
        ...,
        '--credential-id',
        '-i',
        help="Credential UUID to access the cloud storage."
    ),
    metadata_file: Optional[typer.FileText] = typer.Option(
        None,
        '--metadata-file',
        '-f',
        help="Path to file containing the metadata describing your dataset."
             "The metadata should be written in YAML format."
    )
):
    """Link user's own cloud storage to PeriFlow datastore.
    Use `pf datastore --interactive create` command to create a dataset with interactive prompt.
    """
    credential_client: CredentialClientService = build_client(ServiceType.CREDENTIAL)
    credential = credential_client.get_credential(credential_id)
    if credential["type"] != cred_type_map[cloud]:
        secho_error_and_exit(
            f"Credential type and cloud vendor mismatch: {cred_type_map_inv[credential['type']]} and {cloud}."
        )

    storage_helper = build_storage_helper(cloud, credential['value'])
    if storage_path is not None:
        storage_path = storage_path.strip("/")
    files = storage_helper.list_storage_files(storage_name, storage_path)
    if storage_path is not None:
        storage_name = f"{storage_name}/{storage_path}"

    metadata = {}
    if metadata_file is not None:
        try:
            metadata = yaml.safe_load(metadata_file)
        except yaml.YAMLError as exc:
            secho_error_and_exit(f"Error occurred while parsing metadata file... {exc}")

    client: ProjectDataClientService = build_client(ServiceType.PROJECT_DATA)
    datastore = client.create_datastore(name, cloud, region, storage_name, credential_id, metadata, files, True)
    datastore['vendor'] = storage_type_map_inv[datastore['vendor']].value

    typer.secho(f"Datastore ({name}) is created successfully!", fg=typer.colors.BLUE)
    panel_formatter.render([datastore], show_detail=True)
    tree_formatter.render(datastore['files'])
    json_formatter.render(datastore['metadata'])


@app.command()
def upload(
    name: str = typer.Option(
        ...,
        '--name',
        '-n',
        help="Name of your datastore to upload objects."
    ),
    source_path: str = typer.Option(
        ...,
        '--source-path',
        '-p',
        help="Path to source file or directory to upload."
    ),
    metadata_file: Optional[typer.FileText] = typer.Option(
        None,
        '--metadata-file',
        '-f',
        help="Path to file containing the metadata describing your dataset."
             "The metadata should be written in YAML format."
    )
):
    """Create a dataset by uploading dataset files in my local file system.
    The created dataset will have "fai" cloud type.
    """
    client: DataClientService = build_client(ServiceType.DATA)
    project_client: ProjectDataClientService = build_client(ServiceType.PROJECT_DATA)
    expand = source_path.endswith('/')
    source_path: Path = Path(source_path)

    datastore_id = project_client.get_id_by_name(name)
    if datastore_id is not None:
        secho_error_and_exit(f"The datastore with the same name ({name}) already exists.")

    typer.echo(f"Creating datastore ({name})...")
    metadata = {}
    if metadata_file is not None:
        try:
            metadata = yaml.safe_load(metadata_file)
        except yaml.YAMLError as exc:
            secho_error_and_exit(f"Error occurred while parsing metadata file... {exc}")

    datastore = project_client.create_datastore(name, StorageType.FAI, '', '', None, metadata, [], False)
    typer.secho(f"Datastore ({name}) is created successfully.", fg=typer.colors.BLUE)
    datastore_id = datastore['id']

    typer.echo(f"Start uploading objects to datastore ({name})...")
    url_dicts = client.get_upload_urls(
        datastore_id=datastore_id,
        src_path=source_path,
        expand=expand
    )
    upload_files(url_dicts, source_path, expand)

    datastore = client.update_datastore(
        datastore_id,
        files=[
            get_file_info(url_info['path'], source_path, expand) for url_info in url_dicts
        ],
        metadata=metadata,
        active=True
    )
    datastore['vendor'] = storage_type_map_inv[datastore['vendor']].value

    typer.secho(f"Objects are uploaded to datastore ({name}) successfully!", fg=typer.colors.BLUE)
    panel_formatter.render([datastore], show_detail=True)
    tree_formatter.render(datastore['files'])
    json_formatter.render(datastore['metadata'])


@app.command()
def edit(
    name: str = typer.Argument(
        ...,
        help='The name of datastore to update.'
    ),
    new_name: Optional[str] = typer.Option(
        None,
        '--name',
        '-n',
        help='The new name of datastore.'
    ),
    metadata_file: Optional[typer.FileText] = typer.Option(
        None,
        '--metadata-file',
        '-f',
        help="Path to file containing the metadata describing your dataset. "
             "The metadata should be written in YAML format."
    )
):
    """Edit metadata of dataset.
    """
    project_client: ProjectDataClientService = build_client(ServiceType.PROJECT_DATA)
    datastore_id = project_client.get_id_by_name(name)

    metadata = None
    if metadata_file is not None:
        try:
            metadata = yaml.safe_load(metadata_file)
            metadata = metadata or {}
        except yaml.YAMLError as exc:
            secho_error_and_exit(f"Error occurred while parsing metadata file... {exc}")

    client: DataClientService = build_client(ServiceType.DATA)
    datastore = client.update_datastore(
        datastore_id,
        name=new_name,
        metadata=metadata,
        active=True
    )
    datastore['vendor'] = storage_type_map_inv[datastore['vendor']].value

    typer.secho("Datastore is updated successfully!", fg=typer.colors.BLUE)
    panel_formatter.render([datastore], show_detail=True)
    tree_formatter.render(datastore['files'])
    json_formatter.render(datastore['metadata'])


@app.command()
def delete(
    name: str = typer.Argument(
        ...,
        help="Name of datastore to delete.",
    ),
    force: bool = typer.Option(
        False,
        '--force',
        '-f',
        help="Forcefully delete datastore without confirmation prompt."
    )
):
    """Delete dataset from datastore.
    If the dataset was linked from user's cloud storage by `pf datastore link` command, then the storage will not be
    deleted and only the DB record is deleted. If the dataset was uploaded by `pf datastore upload` command, the storage
    will be deleted along with the DB record.
    """
    if not force:
        do_delete = typer.confirm("Are your sure to delete datastore?")
        if not do_delete:
            raise typer.Abort()

    client: DataClientService = build_client(ServiceType.DATA)
    project_client: ProjectDataClientService = build_client(ServiceType.PROJECT_DATA)

    datastore_id = project_client.get_id_by_name(name)
    if datastore_id is None:
        secho_error_and_exit(f"Datastore ({name}) is not found.")

    client.delete_datastore(datastore_id)

    typer.secho(f"Datastore ({name}) deleted successfully!", fg=typer.colors.BLUE)



# Copyright (C) 2021 FriendliAI

from enum import Enum
from typing import Dict


class ServiceType(str, Enum):
    SIGNUP = "SIGNUP"
    USER = "USER"
    USER_GROUP = "USER_GROUP"
    USER_GROUP_PROJECT = "USER_GROUP_PROJECT"
    PROJECT = "PROJECT"
    GROUP = "GROUP"
    GROUP_PROJECT = "GROUP_PROJECT"
    EXPERIMENT = "EXPERIMENT"
    PROJECT_EXPERIMENT = "PROJECT_EXPERIMENT"
    JOB = "JOB"
    JOB_CHECKPOINT = "JOB_CHECKPOINT"
    JOB_ARTIFACT = "JOB_ARTIFACT"
    PROJECT_JOB = "PROJECT_JOB"
    JOB_TEMPLATE = "JOB_TEMPLATE"
    PROJECT_VM_QUOTA = "PROJECT_VM_QUOTA"
    PROJECT_VM_CONFIG = "PROJECT_VM_CONFIG"
    GROUP_VM_CONFIG = "GROUP_VM_CONFIG"
    CREDENTIAL = "CREDENTIAL"
    PROJECT_CREDENTIAL = "PROJECT_CREDENTIAL"
    CREDENTIAL_TYPE = "CREDENTIAL_TYPE"
    DATA = "DATA"
    PROJECT_DATA = "PROJECT_DATA"
    CHECKPOINT = "CHECKPOINT"
    GROUP_PROJECT_CHECKPOINT = "GROUP_PROJECT_CHECKPOINT"
    JOB_WS = "JOB_WS"
    SERVE = "SERVE"
    BILLING_SUMMARY = "BILLING_SUMMARY"


class GroupRole(str, Enum):
    OWNER = 'owner'
    MEMBER = 'member'


class ProjectRole(str, Enum):
    ADMIN = 'admin'
    MAINTAIN = 'maintain'
    DEVELOP = 'develop'
    GUEST = 'guest'


class JobType(str, Enum):
    PREDEFINED = "predefined"
    CUSTOM = "custom"


class LockType(str, Enum):
    ACTIVE = "active"
    STALE = "stale"


class LogType(str, Enum):
    STDOUT = "stdout"
    STDERR = "stderr"
    VMLOG = "vmlog"


class CloudType(str, Enum):
    AWS = "aws"
    AZURE = "azure"
    GCP = "gcp"


class StorageType(str, Enum):
    S3 = "s3"
    BLOB = "azure-blob"
    GCS = "gcs"
    FAI = "fai"


class ModelFormCategory(str, Enum):
    MEGATRON = "MEGATRON"
    ORCA = "ORCA"
    HF = "HF"
    ETC = "ETC"


class GpuType(str, Enum):
    T4 = "t4"
    V100 = "v100"


storage_type_map: Dict[StorageType, str] = {
    StorageType.S3: "aws",
    StorageType.BLOB: "azure.blob",
    StorageType.GCS: "gcp",
    StorageType.FAI: "fai",
}


storage_type_map_inv: Dict[StorageType, str] = {
    "aws": StorageType.S3,
    "azure.blob": StorageType.BLOB,
    "gcp": StorageType.GCS,
    "fai": StorageType.FAI,
}


class CheckpointCategory(str, Enum):
    USER_PROVIDED = "USER"
    JOB_GENERATED = "JOB"


class CredType(str, Enum):
    DOCKER = "docker"
    S3 = "s3"
    BLOB = "azure-blob"
    GCS = "gcs"
    WANDB = "wandb"
    SLACK = "slack"


cred_type_map: Dict[CredType, str] = {
    CredType.DOCKER: "docker",
    CredType.S3: "aws",
    CredType.BLOB: "azure.blob",
    CredType.GCS: "gcp",
    CredType.WANDB: "wandb",
    CredType.SLACK: "slack",
}


cred_type_map_inv: Dict[CredType, str] = {
    "docker": CredType.DOCKER,
    "aws": CredType.S3,
    "azure.blob": CredType.BLOB,
    "gcp": CredType.GCS,
    "wandb": CredType.WANDB,
    "slack": CredType.SLACK,
}

GCP_REGION_NAMES = [
    "asia-east1-a",
    "asia-east1-b",
    "asia-east1-c",
    "asia-east2-a",
    "asia-east2-b",
    "asia-east2-c",
    "asia-northeast1-a",
    "asia-northeast1-b",
    "asia-northeast1-c",
    "asia-northeast2-a",
    "asia-northeast2-b",
    "asia-northeast2-c",
    "asia-northeast3-a",
    "asia-northeast3-b",
    "asia-northeast3-c",
    "asia-south1-a",
    "asia-south1-b",
    "asia-south1-c",
    "asia-south2-a",
    "asia-south2-b",
    "asia-south2-c",
    "asia-southeast1-a",
    "asia-southeast1-b",
    "asia-southeast1-c",
    "asia-southeast2-a",
    "asia-southeast2-b",
    "asia-southeast2-c",
    "australia-southeast1-a",
    "australia-southeast1-b",
    "australia-southeast1-c",
    "australia-southeast2-a",
    "australia-southeast2-b",
    "australia-southeast2-c",
    "europe-central2-a",
    "europe-central2-b",
    "europe-central2-c",
    "europe-north1-a",
    "europe-north1-b",
    "europe-north1-c",
    "europe-west1-b",
    "europe-west1-c",
    "europe-west1-d",
    "europe-west2-a",
    "europe-west2-b",
    "europe-west2-c",
    "europe-west3-a",
    "europe-west3-b",
    "europe-west3-c",
    "europe-west4-a",
    "europe-west4-b",
    "europe-west4-c",
    "europe-west6-a",
    "europe-west6-b",
    "europe-west6-c",
    "northamerica-northeast1-a",
    "northamerica-northeast1-b",
    "northamerica-northeast1-c",
    "northamerica-northeast2-a",
    "northamerica-northeast2-b",
    "northamerica-northeast2-c",
    "southamerica-east1-a",
    "southamerica-east1-b",
    "southamerica-east1-c",
    "southamerica-west1-a",
    "southamerica-west1-b",
    "southamerica-west1-c",
    "us-central1-a",
    "us-central1-b",
    "us-central1-c",
    "us-central1-f",
    "us-east1-b",
    "us-east1-c",
    "us-east1-d",
    "us-east4-a",
    "us-east4-b",
    "us-east4-c",
    "us-west1-a",
    "us-west1-b",
    "us-west1-c",
    "us-west2-a",
    "us-west2-b",
    "us-west2-c",
    "us-west3-a",
    "us-west3-b",
    "us-west3-c",
    "us-west4-a",
    "us-west4-b",
]

AZURE_REGION_NAMES = [
    'eastus',
    'eastus2',
    'southcentralus',
    'westus2',
    'westus3',
    'australiaeast',
    'southeastasia',
    'northeurope',
    'swedencentral',
    'uksouth',
    'westeurope',
    'centralus',
    'northcentralus',
    'westus',
    'southafricanorth',
    'centralindia',
    'eastasia',
    'japaneast',
    'jioindiawest',
    'koreacentral',
    'canadacentral',
    'francecentral',
    'germanywestcentral',
    'norwayeast',
    'switzerlandnorth',
    'uaenorth',
    'brazilsouth',
    'centralusstage'
]

AWS_REGION_NAMES = [
    "us-east-1",
    "us-east-2",
    "us-west-1",
    "us-west-2",
    "eu-west-1",
    "eu-central-1",
    "ap-northeast-1",
    "ap-northeast-2",
    "ap-southeast-1",
    "ap-southeast-2",
    "ap-south-1",
    "sa-east-1",
]

FAI_REGION_NAMES = [
    '',
]

storage_region_map = {
    StorageType.S3: AWS_REGION_NAMES,
    StorageType.BLOB: AZURE_REGION_NAMES,
    StorageType.GCS: GCP_REGION_NAMES,
    StorageType.FAI: FAI_REGION_NAMES,
}


cloud_region_map = {
    CloudType.AWS: AWS_REGION_NAMES,
    CloudType.AZURE: AZURE_REGION_NAMES,
    CloudType.GCP: GCP_REGION_NAMES,
}



# Copyright (C) 2022 FriendliAI

"""PeriFlow Project CLI"""

import uuid
from typing import List, Optional, Tuple, Union

import typer

from pfcli.context import get_current_project_id, set_current_project_id, project_context_path
from pfcli.group import _get_current_org, _get_org_user_id_by_name
from pfcli.service import GroupRole, ProjectRole, ServiceType
from pfcli.service.client import (
    GroupProjectClientService,
    ProjectClientService,
    UserClientService,
    UserGroupProjectClientService,
    build_client
)
from pfcli.service.formatter import PanelFormatter, TableFormatter
from pfcli.utils import secho_error_and_exit


app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False
)
project_table_formatter = TableFormatter(
    name="Project",
    fields=['id', 'name'],
    headers=["ID", "Name"]
)
project_panel_formatter = PanelFormatter(
    name="Project Detail",
    fields=['pf_group_id', 'id', 'name'],
    headers=['Organization ID', 'Project ID', 'Name']
)
member_table_formatter = TableFormatter(
    name="Members",
    fields=["id", "username", "name", "email", "access_level"],
    headers=["ID", "Username", "Name", "Email", "Role"]
)


def _find_project_id(projects: List[dict], project_name: str) -> uuid.UUID:
    for project in projects:
        if project['name'] == project_name:
            return uuid.UUID(project['id'])
    secho_error_and_exit(f"No project exists with name {project_name}.")


@app.command(help="list all accessible projects")
def list(
    tail: Optional[int] = typer.Option(
        None,
        "--tail",
        help="The number of project list to view at the tail"
    ),
    head: Optional[int] = typer.Option(
        None,
        "--tail",
        help="The number of project list to view at the head"
    ),
    show_group_project: bool = typer.Option(
        False,
        "--group",
        "-g",
        help="Show all projects in the current group"
    )
):
    client: Union[GroupProjectClientService, UserGroupProjectClientService]
    if show_group_project:
        client = build_client(ServiceType.GROUP_PROJECT)
    else:
        client = build_client(ServiceType.USER_GROUP_PROJECT)

    projects = client.list_projects()

    if tail is not None or head is not None:
        target_project_list = []
        if tail is not None:
            target_project_list.extend(projects[:tail])
        if head is not None:
            target_project_list.extend(projects[-head:])
    else:
        target_project_list = projects

    project_table_formatter.render(target_project_list)


@app.command(help="create a new project")
def create(
    name: str = typer.Argument(
        ...,
        help='Name of project to create'
    )
):
    client: GroupProjectClientService = build_client(ServiceType.GROUP_PROJECT)
    project_detail = client.create_project(name)
    project_panel_formatter.render(project_detail)


@app.command(help="get current working project")
def current():
    client: ProjectClientService = build_client(ServiceType.PROJECT)
    project_id = get_current_project_id()
    if project_id is None:
        secho_error_and_exit("working project is not set")
    project = client.get_project(project_id)
    project_panel_formatter.render(project)


@app.command(help="switch working project")
def switch(
    name: str = typer.Argument(
        ...,
        help='Name of project to switch',
    )
):
    user_group_project_client: UserGroupProjectClientService = build_client(ServiceType.USER_GROUP_PROJECT)
    project_client: ProjectClientService = build_client(ServiceType.PROJECT)

    project_id = _find_project_id(user_group_project_client.list_projects(), name)
    if project_client.check_project_membership(pf_project_id=project_id):
        set_current_project_id(project_id)
        typer.secho(f"Project switched to {name}.", fg=typer.colors.BLUE)
    else:
        secho_error_and_exit(f"You don't have permission to project ({name}). Please contact to the project admin.")


@app.command(help="delete project")
def delete(
    name: str = typer.Argument(
        ...,
        help="Name of project to delete",
    )
):
    project_client: ProjectClientService = build_client(ServiceType.PROJECT)
    user_group_project_client: UserGroupProjectClientService = build_client(ServiceType.USER_GROUP_PROJECT)
    project_id = _find_project_id(user_group_project_client.list_projects(), name)
    project_client.delete_project(pf_project_id=project_id)
    if project_id == get_current_project_id():
        project_context_path.unlink()
    typer.secho(f"Project {name} deleted.", fg=typer.colors.BLUE)


@app.command("add-user", help="add user to project")
def add_user(
    username: str = typer.Argument(
        ...,
        help="Username to add to the current working project",
    ),
    role: ProjectRole = typer.Argument(
        ...,
        help="Project role to assign",
    )
):
    user_client: UserClientService = build_client(ServiceType.USER)

    org_id, project_id = _check_project_and_get_id()
    user_id = _get_org_user_id_by_name(org_id, username)

    user_client.add_to_project(user_id, project_id, role)
    typer.secho(f"User successfully added to project")


@app.command("set-role", help="set project role for a user")
def set_role(
    username: str = typer.Argument(
        ...,
        help="Username to set project role",
    ),
    role: ProjectRole = typer.Argument(
        ...,
        help="Project role",
    )
):
    user_client: UserClientService = build_client(ServiceType.USER)

    org_id, project_id = _check_project_and_get_id()
    user_id = _get_org_user_id_by_name(org_id, username)

    user_client.set_project_privilege(user_id, project_id, role)
    typer.secho(f"Project role for user ({username}) successfully updated to {role.value}!")


@app.command(help="list up members in the current working project")
def members():
    project_client: ProjectClientService = build_client(ServiceType.PROJECT)

    project_id = get_current_project_id()
    members = project_client.list_users(project_id)
    member_table_formatter.render(members)


def _check_project_and_get_id() -> Tuple[str, str]:
    """Get org_id and project_id if valid"""

    user_client: UserClientService = build_client(ServiceType.USER)

    org = _get_current_org()
    project_id = get_current_project_id()

    if org['privilege_level'] == GroupRole.OWNER:
        return org['id'], project_id

    requester = user_client.get_project_membership(project_id)
    if requester['access_level'] != ProjectRole.ADMIN:
        secho_error_and_exit("Only the admin of the project can add-user/set-role")

    return org['id'], project_id



# Copyright (C) 2022 FriendliAI

"""PeriFlow ServeClient Service"""

from typing import List

from pfcli.service.client.base import ClientService, T, safe_request


class ServeClientService(ClientService):
    def get_serve(self, serve_id: T) -> dict:
        response = safe_request(self.retrieve, err_prefix=f"Serve ({serve_id}) is not found. You may enter wrongID.")(
            pk=serve_id
        )
        return response.json()

    def create_serve(self, config = dict) -> dict:
        response = safe_request(self.post, err_prefix="Failed to post new serve.")(
            json=config
        )
        return response.json()

    def list_serves(self) -> List[dict]:
        response = safe_request(self.list, err_prefix="Failed to list serves.")()
        return response.json()

    def delete_serve(self, serve_id: T) -> None:
        safe_request(self.delete, err_prefix="Failed to delete serve.")(
            pk=serve_id
        )
